#!C:/perl/bin/perl
use POSIX;

$| = 1;

foreach my $state ('al', 'ak', 'az', 'ar', 'ca', 'co', 'ct', 'de', 'fl', 'ga', 'hi', 'id', 'il', 'in', 'ia', 'ks', 'ky', 'la', 'me', 'md', 'ma', 'mi', 'mn', 'ms', 'mo', 'mt', 'ne', 'nv', 'nh', 'nj', 'nm', 'ny', 'nc', 'nd', 'oh', 'ok', 'or', 'pa', 'ri', 'sc', 'sd', 'tn', 'tx', 'ut', 'vt', 'va', 'wa', 'wv', 'wi', 'wy', 'dc') {

print "Working on state '$state'...";

# make sure the county directory exists
if (! -d "$state/counties") {
    print "directory does not exist\n";
    next;
}

`del mapdata/$state/counties/*.xml`;
`del mapdata/$state/counties/*.pt`;
`del mapdata/$state/counties/*.enc`;
next;
# read the key file
open(KEYFILE, "< $state/counties/index.dat");
while (! eof(KEYFILE)) {
    $key = <KEYFILE>;
    $tmp1 = <KEYFILE>;
    $ptcount = <KEYFILE>;
    $county = <KEYFILE>;
    $tmp2 = <KEYFILE>;
    $tmp3 = <KEYFILE>;
    chomp($key);
    chomp($county);
    $key =~ s/^\s+//;
    $county =~ s/^\s+//;
    $county =~ s/"//g;
    if ($key > 0) {
	$index[$key] = lc($county);
    }
    $blank = <KEYFILE>;
}
close(KEYFILE);

open(DATAFILE, "< $state/counties/data.dat");
open(CNTRFILE, "> $state/counties/geocenter.txt");

# set the original values of the lat and long to an invalid number
my $len = 0;
my @points;

my $small0 = 500000;

use constant LAT => 0;
use constant LNG => 1;
use constant LEV => 2;

while (! eof(DATAFILE)) {
    $line = <DATAFILE>;
    chomp($line);

    # three points marks the start of a new county
    if ($line =~ /^\s+-?(\d+)\s+(\S+)\s+(\S+)$/) {
	$key = $1;
	print CNTRFILE "$index[$1] $2 $3\n";
    }

    elsif ($line =~ /^END/) {
	if ($len > 0) {
	    open(PTFILE, "> $state/counties/$index[$key]\.pt");
	    for (my $i = 0; $i < $len; $i++) {
		print PTFILE "$points[$i][LAT] $points[$i][LNG]\n";
	    }
	    close(PTFILE);

	    open(XMLFILE, "> $state/counties/$index[$key]\.xml");
	    print XMLFILE "<boundary>\n";
	    for (my $i = 0; $i < $len; $i++) {
		printf XMLFILE "  <pt lat=\"%.8f\" lng=\"%.8f\" />\n", $points[$i][LAT], $points[$i][LNG];
	    }
	    print XMLFILE "</boundary>";
	    close(XMLFILE);

	    setLevels();
	    my $result = createEncodings();
	    open(ENCPTFILE, "> $state/counties/$index[$key]\.enc");
	    print ENCPTFILE $result->[0]."\n";
	    print ENCPTFILE $result->[1]."\n";
	    close(ENCPTFILE);
	}
	$len = 0;
	@points = ();
    }

    elsif ($line =~ /^\s+(\S+)\s+(\S+)$/) {
	# for the standard point file, just put the coords in the file as
	#    latitutde longitude
	$points[$len] = [$2, $1, "unset"];
	$len++;
    }
    
    else {
	print "ERROR reading line $line\n";
    }
}
close(CNTRFILE);
close(DATAFILE);
print "done\n";
}

sub setLevels {
    my $numLevels = 18;
    my $small;
    my $i;
    my $j;
	
    # Set the endpoints to show at all zoom levels
    $points[0][LEV] = $numLevels - 1;
    $points[$len-1][LEV] = $numLevels-1;

    # Starting at the largest possible level, we step down through
    # the levels labelling points which need to appear at
    # lower resolutions.
    my $level;
    for ($level = $numLevels - 1; $level >= 0; $level--) {
	# Set "small" for the current level.
	# More generally, we might use
	# small = small0*Math.pow(2.0, -18*(numLevels-level)/numLevels);
    	$small = $small0 * (2.0 ** ($level - 18));
    	if ($level == 0) {
	    $small = 0
    	}
    
	# Figure out which points need to occur at this level,
	# the main criteria being that we've stepped a certain 
	# distance (called "small") from the previous set point.
    	$i = 1;
    	$j = 1;
    	while ($i < $len - 1) {
	    if ((dist($points[$i], $points[$j]) >= $small) && ($points[$i][LEV] eq "unset")) {
		$points[$i][LEV] = $level;
		$j = $i;
	    }
	    if ($points[$i][LEV] ne "unset") {
		$j = $i;
	    }
	    $points[$i][LEV] = 'P';
	    $i++;
	}
    }
}

sub dist {
    my $pt1 = shift;
    my $pt2 = shift;
    my $deg = 0.0174532925199;

    return 12746004.5 * asin(sqrt(
  		(sin(($point1[LAT] - $point2[LAT])* $deg/2) ** 2) + 
  		cos($point1[LAT] * $deg) * cos($point2[LAT] * $deg) * 
  		(sin(($point1[LNG] - $point2[LNG])* $deg/2) ** 2)));
}

sub createEncodings {
    my $i = 0;

    my $plat = 0;
    my $plng = 0;

    my $encoded_points = '';
    my $encoded_levels = '';

    for ($i = 0; $i < $len; $i++) {
	my $point = $points[$i];
	my $lat = $point->[LAT];
	my $lng = $point->[LNG];
	my $level = $point->[LEV];

	my $late5 = floor($lat * 1e5);
	my $lnge5 = floor($lng * 1e5);

	my $dlat = $late5 - $plat;
	my $dlng = $lnge5 - $plng;

	$plat = $late5;
	$plng = $lnge5;

	$encoded_points .= encodeSignedNumber($dlat) . encodeSignedNumber($dlng);
	$encoded_levels .= encodeNumber($level);
    }
    return([$encoded_points, $encoded_levels]);
}

sub encodeSignedNumber {
    my $num = shift;
    my $sign_num = $num << 1;

    if ($num < 0) {
	$sign_num = ~($sign_num);
    }
    return(encodeNumber($sign_num));
}

sub encodeNumber {
    my $num = shift;

    my $encode_string = '';
    my $next_val;

    while ($num >= 0x20) {
	$next_val = (0x20 | ($num & 0x1f)) + 63;
	if ($next_val == 92) {
	    $encode_string .= chr($next_val);
	}
	$encode_string .= chr($next_val);
	$num >>= 5;
    }

    my $final_val = $num + 63;
    if ($final_val == 92) {
	$encode_string .= chr($final_val);
    }
    $encode_string .= chr($final_val);
    return($encode_string);
}

